from flask import json


def get_event(entry_path, method, entry_path_parameters=None, params=None, body=None, headers=None):
    headers = headers or {
        'accept': '*/*',
        'authorization': 'token',
        'content-type': 'application/json'
    }
    return {
        'type': 'LAMBDA',
        'resource': entry_path,
        'path': entry_path,
        'httpMethod': method.upper(),
        'headers': {
            'accept-encoding': 'gzip, deflate, br',
            'cache-control': 'max-age=0',
            'cookie': 'session=70b58773-af3e-4153-a5ab-5356481ea87e',
            'host': 'zb53tzphr8.execute-api.eu-west-1.amazonaws.com',
            'origin': 'https://neorezo.io',
            'sec-ch-ua': '" Not;A Brand";v="99", "Google Chrome";v="97", "Chromium";v="97"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Linux"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Amazon CloudFront',
            'via': '2.0 4e4ca876a59e9f2e22ec751bbab5f282.cloudfront.net (CloudFront)',
            'x-amz-cf-id': '0gvkeT2_kuutjZHh075zWXA0pfqm-afERp4UltS3XBAguLOI4tdckw==',
            'x-amzn-trace-id': 'Root=1-620f51ee-3f8b28746271cbea29b0084e',
            'x-forwarded-for': '91.170.42.203, 130.176.185.118',
            'x-forwarded-port': '443',
            'x-forwarded-proto': 'https',
            **headers
        },
        'multiValueHeaders': {},
        'body': json.dumps(body or {}),
        'queryStringParameters': {},
        'multiValueQueryStringParameters': params or {},
        'pathParameters': {},
        'stageVariables': None,
        'isBase64Encoded': False,
        'entryPathParameters': entry_path_parameters or {},
        'requestContext': {
            'httpMethod': method.upper(),
            'resourceId': 'fp5ol74tr7',
            'entryPath': entry_path,
            'extendedRequestId': 'EktgyFweDoEFabw=',
            'requestTime': '24/Aug/2021:13:37:08 +0000',
            'path': f"/dev{entry_path}",
            'accountId': '935392763270',
            'protocol': 'HTTP/1.1',
            'stage': 'dev',
            'domainPrefix': 'htzd2rneg1',
            'requestTimeEpoch': 1629812228818,
            'requestId': '2fa7f00a-58fe-4f46-a829-1fee00898e42',
            'domainName': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com',
            'apiId': 'htzd2rneg1'
        },
        'params': {
            'path': {},
            'querystring': {},
            'header': {'Accept': '*/*', 'Authorization': 'token', 'CloudFront-Forwarded-Proto': 'https',
                       'CloudFront-Is-Desktop-Viewer': 'true', 'CloudFront-Is-Mobile-Viewer': 'false',
                       'CloudFront-Is-SmartTV-Viewer': 'false', 'CloudFront-Is-Tablet-Viewer': 'false',
                       'CloudFront-Viewer-Country': 'FR', 'content-type': 'application/json',
                       'Cookie': 'session=70b58773-af3e-4153-a5ab-5356481ea87e',
                       'Host': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com', 'User-Agent': 'insomnia/2021.4.1',
                       'Via': '2.0 4dd111c814b0b5cf8bf82e59008da625.cloudfront.net (CloudFront)',
                       'X-Amz-Cf-Id': 'ka1hbQCSUOZ-d0VQYuE_gtF4icy443t7kP3UGsDLZDF_5QyTX13FoQ==',
                       'X-Amzn-Trace-Id': 'Root=1-6124f604-3fb9457c7489ebf14ed0f8f6',
                       'X-Forwarded-For': '78.234.174.193, 130.176.152.165', 'X-Forwarded-Port': '443',
                       'X-Forwarded-Proto': 'https'}},
        'context': {
            'resourceId': 'fp5ol74tr7',
            'authorizer': '',
            'resourcePath': '/',
            'httpMethod': 'GET',
            'extendedRequestId': 'EktgyFweDoEFabw=',
            'requestTime': '24/Aug/2021:13:37:08 +0000',
            'path': '/dev/',
            'accountId': '935392763270',
            'protocol': 'HTTP/1.1',
            'requestOverride': '',
            'stage': 'dev',
            'domainPrefix': 'htzd2rneg1',
            'requestTimeEpoch': '1629812228818',
            'requestId': '2fa7f00a-58fe-4f46-a829-1fee00898e42',
            'identity': '',
            'domainName': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com',
            'responseOverride': '',
            'apiId': 'htzd2rneg1'
        }
    }
