{% include 'terraform/provider.j2' with context %}

{% include 'terraform/local.j2' with context %}

{% include 'terraform/role.j2' with context %}

{% include 'terraform/rest_api.j2' with context %}

{% include 'terraform/lambda.j2' with context %}

{% include 'terraform/output.j2' with context %}
